// Copyright (C) 1996 DIMACS Center, Rutgers, The State University of New Jersey
// Author(s): Patricia K. Fasel (Los Alamos Nat. Lab.), Jonathan Berry,
//	      Chris Burrows (Trenton State College)
//	      Jon Berry (Elon College)
// This software is copyrighted by the DIMACS Center at Rutgers, The State
// University of New Jersey.  IT IS PROVIDED AS IS, AND THE AUTHORS, DIMACS, AND
// RUTGERS, THE STATE UNIVERSITY OF NEW JERSEY  DISCLAIM
// ALL LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
// DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
// IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
// NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
// MODIFICATIONS.
// The authors hereby grant permission to use, copy, modify, distribute,
// and license this software and its documentation for any purpose, provided
// that existing copyright notices are retained in all copies and that this
// notice is included verbatim in any distributions. No written agreement,
// license, or royalty fee is required for any of the authorized uses.
// Modifications to this software may be copyrighted by their authors
// and need not follow the licensing terms described here, provided that
// the new terms are clearly indicated on the first page of each file where
// they apply.
// Last File Update: 28-Jun-1999
// 
#include <iostream.h>
#include <LINK/graph/Graph.h>
#include <LINK/graph/UBinGraph.h>
#include <LINK/graph/Vertex.h>
#include <LINK/graph/Attribute.h>
#include <LINK/basic/Sequence.h>
#include <LINK/basic/SetFuncs.h>
#include <LINK/algorithm/Algorithms.h>

int above_in_one(Vertex *w, Vertex *a, Vertex *b, Sequence<Vertex*> sq)
{
	if ((sq.rank(w) < sq.rank(a)) && (sq.rank(w) < sq.rank(b)))
		return 1;
	return 0;
}

int above(Vertex *w, Vertex *a, Vertex *b, Set<Vertex*> st, Set<int> perms)
{
	Iterator<int> next_rank(&perms);
	int rank;
	while (next_rank(rank)) {
		Sequence<Vertex*> next_perm = 
			SetFuncs<Vertex*>::permutation(st, rank);
		if (above_in_one(w,a,b,next_perm)) {
			return 1;
		}
	}
	return 0;
}

int edge_ab(Set<Vertex*> potential_edge, Set<Vertex*> verts, Set<int> perms)
{
	Iterator<Vertex*> get_vert(&verts);
	Vertex *v, *a, *b;
	a = potential_edge.ref(0);
	b = potential_edge.ref(1);
	while (get_vert(v)) {
		if ((v == a) || (v == b))
			continue;
		if (!above(v, a, b, verts, perms))
			return 0;
	}
	return 1;
}

UBinGraph *GenGraphFromLinearOrders(int num, Set<int> perms)
{
	UBinGraph *g = new UBinGraph;
	g->addVertices(num);
	if (g->order() <= 0)
		return g;
	Set<Vertex*> verts = g->vertices();
	Set<Set<Vertex*> > pairs = SetFuncs<Vertex*>::choose(verts,2);
	Iterator<Set<Vertex*> > get_potential_edge(&pairs);
	Set<Vertex*> next_pair;
	while (get_potential_edge(next_pair)) {
		if (edge_ab(next_pair, verts, perms))
			g->addEdge(next_pair);
	}
	return g;
}

void UpdateLinearOrdersGraph(Graph *g, Set<int> perms)
{
	Set<Vertex*> verts = g->vertices();
	Set<Set<Vertex*> > pairs = SetFuncs<Vertex*>::choose(verts,2);
	Iterator<Set<Vertex*> > get_potential_edge(&pairs);
	Set<Vertex*> next_pair;
	Edge *e;
	while (get_potential_edge(next_pair)) {
		if (edge_ab(next_pair, verts, perms))
			g->addEdge(next_pair);
		else if (e=g->isEdge(next_pair)) {
			g->deleteEdge(next_pair);
		}
	}
	return;
}
